'use strict';

Object.defineProperty(exports, "__esModule", {
  value: true
});

var _typeof = typeof Symbol === "function" && typeof Symbol.iterator === "symbol" ? function (obj) { return typeof obj; } : function (obj) { return obj && typeof Symbol === "function" && obj.constructor === Symbol && obj !== Symbol.prototype ? "symbol" : typeof obj; };

exports.default = function (_ref8) {
  var t = _ref8.types,
      template = _ref8.template,
      options = _ref8.options;

  function _ref(_id) {
    if (!Plugin(_id)) {
      throw new TypeError('Function return value violates contract.\n\nExpected:\nPlugin\n\nGot:\n' + _inspect(_id));
    }

    return _id;
  }

  if (!PluginParams(arguments[0])) {
    throw new TypeError('Value of argument 0 violates contract.\n\nExpected:\nPluginParams\n\nGot:\n' + _inspect(arguments[0]));
  }

  var defaultNames = {
    assert: 'assert',
    precondition: 'pre',
    postcondition: 'post',
    invariant: 'invariant',
    return: 'it',
    old: 'old'
  };

  var NAMES = Object.assign({}, defaultNames);

  var guard = template('\n    if (!condition) {\n      throw new Error(message);\n    }\n  ');

  if (!(typeof guard === 'function')) {
    throw new TypeError('Value of variable "guard" violates contract.\n\nExpected:\n({ [key: string]: Node\n}) => Node\n\nGot:\n' + _inspect(guard));
  }

  var guardFn = template('\n    const id = (it) => {\n      conditions;\n      return it;\n    }\n  ');

  if (!(typeof guardFn === 'function')) {
    throw new TypeError('Value of variable "guardFn" violates contract.\n\nExpected:\n({ [key: string]: Node\n}) => Node\n\nGot:\n' + _inspect(guardFn));
  }

  function assemblePrecondition(path) {
    if (!NodePath(path)) {
      throw new TypeError('Value of argument "path" violates contract.\n\nExpected:\nNodePath\n\nGot:\n' + _inspect(path));
    }

    var body = path.get('body');

    if (!NodePath(body)) {
      throw new TypeError('Value of variable "body" violates contract.\n\nExpected:\nNodePath\n\nGot:\n' + _inspect(body));
    }

    var fn = path.getFunctionParent();

    if (!NodePath(fn)) {
      throw new TypeError('Value of variable "fn" violates contract.\n\nExpected:\nNodePath\n\nGot:\n' + _inspect(fn));
    }

    var name = fn.node.id ? '"' + fn.node.id.name + '" ' : ' ';
    if (body.isExpressionStatement()) {
      var condition = body.get('expression');

      if (!NodePath(condition)) {
        throw new TypeError('Value of variable "condition" violates contract.\n\nExpected:\nNodePath\n\nGot:\n' + _inspect(condition));
      }

      var message = void 0;

      if (!(message == null || Node(message))) {
        throw new TypeError('Value of variable "message" violates contract.\n\nExpected:\n?Node\n\nGot:\n' + _inspect(message));
      }

      if (condition.isSequenceExpression()) {
        var expressions = condition.get('expressions');
        condition = expressions[0];

        if (!NodePath(condition)) {
          throw new TypeError('Value of variable "condition" violates contract.\n\nExpected:\nNodePath\n\nGot:\n' + _inspect(condition));
        }

        message = expressions[1].node;

        if (!(message == null || Node(message))) {
          throw new TypeError('Value of variable "message" violates contract.\n\nExpected:\n?Node\n\nGot:\n' + _inspect(message));
        }
      } else {
        message = t.stringLiteral('Function ' + name + 'precondition failed: ' + (0, _babelGenerator2.default)(condition.node).code);

        if (!(message == null || Node(message))) {
          throw new TypeError('Value of variable "message" violates contract.\n\nExpected:\n?Node\n\nGot:\n' + _inspect(message));
        }
      }
      path.replaceWith(guard({
        condition: staticCheck(condition),
        message: message
      }));
      return;
    }

    body.traverse({
      "VariableDeclaration|Function|AssignmentExpression|UpdateExpression|YieldExpression|ReturnStatement": function VariableDeclarationFunctionAssignmentExpressionUpdateExpressionYieldExpressionReturnStatement(item) {
        if (!NodePath(item)) {
          throw new TypeError('Value of argument "item" violates contract.\n\nExpected:\nNodePath\n\nGot:\n' + _inspect(item));
        }

        throw path.buildCodeFrameError('Preconditions cannot have side effects.');
      },
      ExpressionStatement: function ExpressionStatement(statement) {
        if (!NodePath(statement)) {
          throw new TypeError('Value of argument "statement" violates contract.\n\nExpected:\nNodePath\n\nGot:\n' + _inspect(statement));
        }

        var condition = statement.get('expression');

        if (!NodePath(condition)) {
          throw new TypeError('Value of variable "condition" violates contract.\n\nExpected:\nNodePath\n\nGot:\n' + _inspect(condition));
        }

        var message = void 0;

        if (!(message == null || Node(message))) {
          throw new TypeError('Value of variable "message" violates contract.\n\nExpected:\n?Node\n\nGot:\n' + _inspect(message));
        }

        if (condition.isSequenceExpression()) {
          var _expressions = condition.get('expressions');
          condition = _expressions[0];

          if (!NodePath(condition)) {
            throw new TypeError('Value of variable "condition" violates contract.\n\nExpected:\nNodePath\n\nGot:\n' + _inspect(condition));
          }

          message = _expressions[1].node;

          if (!(message == null || Node(message))) {
            throw new TypeError('Value of variable "message" violates contract.\n\nExpected:\n?Node\n\nGot:\n' + _inspect(message));
          }
        } else {
          message = t.stringLiteral('Function ' + name + 'precondition failed: ' + (0, _babelGenerator2.default)(condition.node).code);

          if (!(message == null || Node(message))) {
            throw new TypeError('Value of variable "message" violates contract.\n\nExpected:\n?Node\n\nGot:\n' + _inspect(message));
          }
        }
        statement.replaceWith(guard({
          condition: staticCheck(condition),
          message: message
        }));
      }
    });

    if (body.isBlockStatement()) {
      path.replaceWithMultiple(path.get('body').node.body);
    } else {
      path.replaceWith(path.get('body'));
    }
  }

  function assemblePostcondition(path) {
    function _ref3(_id5) {
      if (!Identifier(_id5)) {
        throw new TypeError('Function "assemblePostcondition" return value violates contract.\n\nExpected:\nIdentifier\n\nGot:\n' + _inspect(_id5));
      }

      return _id5;
    }

    if (!NodePath(path)) {
      throw new TypeError('Value of argument "path" violates contract.\n\nExpected:\nNodePath\n\nGot:\n' + _inspect(path));
    }

    var body = path.get('body');

    if (!NodePath(body)) {
      throw new TypeError('Value of variable "body" violates contract.\n\nExpected:\nNodePath\n\nGot:\n' + _inspect(body));
    }

    var fn = path.getFunctionParent();

    if (!NodePath(fn)) {
      throw new TypeError('Value of variable "fn" violates contract.\n\nExpected:\nNodePath\n\nGot:\n' + _inspect(fn));
    }

    var name = fn.node.id ? '"' + fn.node.id.name + '" ' : ' ';
    var conditions = [];

    if (!(Array.isArray(conditions) && conditions.every(function (item) {
      return Node(item);
    }))) {
      throw new TypeError('Value of variable "conditions" violates contract.\n\nExpected:\nNode[]\n\nGot:\n' + _inspect(conditions));
    }

    var captures = [];

    if (!(Array.isArray(captures) && captures.every(function (item) {
      return Node(item);
    }))) {
      throw new TypeError('Value of variable "captures" violates contract.\n\nExpected:\nNode[]\n\nGot:\n' + _inspect(captures));
    }

    if (body.isExpressionStatement()) {
      var condition = body.get('expression');

      if (!NodePath(condition)) {
        throw new TypeError('Value of variable "condition" violates contract.\n\nExpected:\nNodePath\n\nGot:\n' + _inspect(condition));
      }

      var message = void 0;

      if (!(message == null || Node(message))) {
        throw new TypeError('Value of variable "message" violates contract.\n\nExpected:\n?Node\n\nGot:\n' + _inspect(message));
      }

      if (condition.isSequenceExpression()) {
        var expressions = condition.get('expressions');
        condition = expressions[0];

        if (!NodePath(condition)) {
          throw new TypeError('Value of variable "condition" violates contract.\n\nExpected:\nNodePath\n\nGot:\n' + _inspect(condition));
        }

        message = expressions[1].node;

        if (!(message == null || Node(message))) {
          throw new TypeError('Value of variable "message" violates contract.\n\nExpected:\n?Node\n\nGot:\n' + _inspect(message));
        }
      } else {
        message = t.stringLiteral('Function ' + name + 'postcondition failed: ' + (0, _babelGenerator2.default)(condition.node).code);

        if (!(message == null || Node(message))) {
          throw new TypeError('Value of variable "message" violates contract.\n\nExpected:\n?Node\n\nGot:\n' + _inspect(message));
        }
      }
      conditions.push(guard({
        condition: staticCheck(condition),
        message: message
      }));
    } else {
      body.traverse({
        "VariableDeclaration|Function|AssignmentExpression|UpdateExpression|YieldExpression|ReturnStatement": function VariableDeclarationFunctionAssignmentExpressionUpdateExpressionYieldExpressionReturnStatement(item) {
          if (!NodePath(item)) {
            throw new TypeError('Value of argument "item" violates contract.\n\nExpected:\nNodePath\n\nGot:\n' + _inspect(item));
          }

          throw path.buildCodeFrameError('Postconditions cannot have side effects.');
        },
        CallExpression: function CallExpression(call) {
          if (!NodePath(call)) {
            throw new TypeError('Value of argument "call" violates contract.\n\nExpected:\nNodePath\n\nGot:\n' + _inspect(call));
          }

          var callee = call.get('callee');

          if (!NodePath(callee)) {
            throw new TypeError('Value of variable "callee" violates contract.\n\nExpected:\nNodePath\n\nGot:\n' + _inspect(callee));
          }

          var args = call.get('arguments');

          if (!(Array.isArray(args) && args.every(function (item) {
            return NodePath(item);
          }))) {
            throw new TypeError('Value of variable "args" violates contract.\n\nExpected:\nNodePath[]\n\nGot:\n' + _inspect(args));
          }

          if (!callee.isIdentifier() || callee.node.name !== NAMES.old || call.scope.hasBinding(NAMES.old) || args.length === 0) {
            return;
          }
          var argument = args[0];

          if (!NodePath(argument)) {
            throw new TypeError('Value of variable "argument" violates contract.\n\nExpected:\nNodePath\n\nGot:\n' + _inspect(argument));
          }

          var id = call.scope.generateUidIdentifierBasedOnNode(argument.node);
          fn.scope.push({ id: id, init: argument.node, kind: 'const' });
          call.replaceWith(id);
        },
        ExpressionStatement: function ExpressionStatement(statement) {
          if (!NodePath(statement)) {
            throw new TypeError('Value of argument "statement" violates contract.\n\nExpected:\nNodePath\n\nGot:\n' + _inspect(statement));
          }

          var condition = statement.get('expression');

          if (!NodePath(condition)) {
            throw new TypeError('Value of variable "condition" violates contract.\n\nExpected:\nNodePath\n\nGot:\n' + _inspect(condition));
          }

          var message = void 0;

          if (!(message == null || Node(message))) {
            throw new TypeError('Value of variable "message" violates contract.\n\nExpected:\n?Node\n\nGot:\n' + _inspect(message));
          }

          if (condition.isSequenceExpression()) {
            var _expressions2 = condition.get('expressions');
            condition = _expressions2[0];

            if (!NodePath(condition)) {
              throw new TypeError('Value of variable "condition" violates contract.\n\nExpected:\nNodePath\n\nGot:\n' + _inspect(condition));
            }

            message = _expressions2[1].node;

            if (!(message == null || Node(message))) {
              throw new TypeError('Value of variable "message" violates contract.\n\nExpected:\n?Node\n\nGot:\n' + _inspect(message));
            }
          } else {
            message = t.stringLiteral('Function ' + name + 'postcondition failed: ' + (0, _babelGenerator2.default)(condition.node).code);

            if (!(message == null || Node(message))) {
              throw new TypeError('Value of variable "message" violates contract.\n\nExpected:\n?Node\n\nGot:\n' + _inspect(message));
            }
          }
          statement.replaceWith(guard({
            condition: staticCheck(condition),
            message: message
          }));
        }
      });
      conditions.push.apply(conditions, _toConsumableArray(body.node.body));
    }

    var id = path.scope.generateUidIdentifier((fn.node.id ? fn.node.id.name : 'check') + 'Postcondition');

    fn.get('body').get('body')[0].insertBefore(guardFn({
      id: id,
      conditions: conditions,
      it: t.identifier(NAMES.return)
    }));

    path.remove();
    return _ref3(id);
  }

  function assembleAssertion(path) {
    if (!NodePath(path)) {
      throw new TypeError('Value of argument "path" violates contract.\n\nExpected:\nNodePath\n\nGot:\n' + _inspect(path));
    }

    var body = path.get('body');

    if (!NodePath(body)) {
      throw new TypeError('Value of variable "body" violates contract.\n\nExpected:\nNodePath\n\nGot:\n' + _inspect(body));
    }

    var fn = path.getFunctionParent();

    if (!(fn == null || NodePath(fn))) {
      throw new TypeError('Value of variable "fn" violates contract.\n\nExpected:\n?NodePath\n\nGot:\n' + _inspect(fn));
    }

    var name = fn && fn.node && fn.node.id ? '"' + fn.node.id.name + '"' : '';
    if (body.isExpressionStatement()) {
      var condition = body.get('expression');

      if (!NodePath(condition)) {
        throw new TypeError('Value of variable "condition" violates contract.\n\nExpected:\nNodePath\n\nGot:\n' + _inspect(condition));
      }

      var message = void 0;

      if (!(message == null || Node(message))) {
        throw new TypeError('Value of variable "message" violates contract.\n\nExpected:\n?Node\n\nGot:\n' + _inspect(message));
      }

      if (condition.isSequenceExpression()) {
        var expressions = condition.get('expressions');
        condition = expressions[0];

        if (!NodePath(condition)) {
          throw new TypeError('Value of variable "condition" violates contract.\n\nExpected:\nNodePath\n\nGot:\n' + _inspect(condition));
        }

        message = expressions[1].node;

        if (!(message == null || Node(message))) {
          throw new TypeError('Value of variable "message" violates contract.\n\nExpected:\n?Node\n\nGot:\n' + _inspect(message));
        }
      } else if (name) {
        message = t.stringLiteral('Function ' + name + ' assertion failed: ' + (0, _babelGenerator2.default)(condition.node).code);

        if (!(message == null || Node(message))) {
          throw new TypeError('Value of variable "message" violates contract.\n\nExpected:\n?Node\n\nGot:\n' + _inspect(message));
        }
      } else {
        message = t.stringLiteral('Assertion failed: ' + (0, _babelGenerator2.default)(condition.node).code);

        if (!(message == null || Node(message))) {
          throw new TypeError('Value of variable "message" violates contract.\n\nExpected:\n?Node\n\nGot:\n' + _inspect(message));
        }
      }
      path.replaceWith(guard({
        condition: staticCheck(condition),
        message: message
      }));
      return;
    }

    body.traverse({
      "VariableDeclaration|Function|AssignmentExpression|UpdateExpression|YieldExpression|ReturnStatement": function VariableDeclarationFunctionAssignmentExpressionUpdateExpressionYieldExpressionReturnStatement(item) {
        if (!NodePath(item)) {
          throw new TypeError('Value of argument "item" violates contract.\n\nExpected:\nNodePath\n\nGot:\n' + _inspect(item));
        }

        throw path.buildCodeFrameError('Assertions cannot have side effects.');
      },
      ExpressionStatement: function ExpressionStatement(statement) {
        if (!NodePath(statement)) {
          throw new TypeError('Value of argument "statement" violates contract.\n\nExpected:\nNodePath\n\nGot:\n' + _inspect(statement));
        }

        var condition = statement.get('expression');

        if (!NodePath(condition)) {
          throw new TypeError('Value of variable "condition" violates contract.\n\nExpected:\nNodePath\n\nGot:\n' + _inspect(condition));
        }

        var message = void 0;

        if (!(message == null || Node(message))) {
          throw new TypeError('Value of variable "message" violates contract.\n\nExpected:\n?Node\n\nGot:\n' + _inspect(message));
        }

        if (condition.isSequenceExpression()) {
          var _expressions3 = condition.get('expressions');
          condition = _expressions3[0];

          if (!NodePath(condition)) {
            throw new TypeError('Value of variable "condition" violates contract.\n\nExpected:\nNodePath\n\nGot:\n' + _inspect(condition));
          }

          message = _expressions3[1].node;

          if (!(message == null || Node(message))) {
            throw new TypeError('Value of variable "message" violates contract.\n\nExpected:\n?Node\n\nGot:\n' + _inspect(message));
          }
        } else if (name) {
          message = t.stringLiteral('Function ' + name + ' assertion failed: ' + (0, _babelGenerator2.default)(condition.node).code);

          if (!(message == null || Node(message))) {
            throw new TypeError('Value of variable "message" violates contract.\n\nExpected:\n?Node\n\nGot:\n' + _inspect(message));
          }
        } else {
          message = t.stringLiteral('Assertion failed: ' + (0, _babelGenerator2.default)(condition.node).code);

          if (!(message == null || Node(message))) {
            throw new TypeError('Value of variable "message" violates contract.\n\nExpected:\n?Node\n\nGot:\n' + _inspect(message));
          }
        }
        statement.replaceWith(guard({
          condition: staticCheck(condition),
          message: message
        }));
      }
    });

    if (body.isBlockStatement()) {
      path.replaceWithMultiple(path.get('body').node.body);
    } else {
      path.replaceWith(path.get('body'));
    }
  }

  function assembleInvariant(path) {
    function _ref5(_id12) {
      if (!Identifier(_id12)) {
        throw new TypeError('Function "assembleInvariant" return value violates contract.\n\nExpected:\nIdentifier\n\nGot:\n' + _inspect(_id12));
      }

      return _id12;
    }

    if (!NodePath(path)) {
      throw new TypeError('Value of argument "path" violates contract.\n\nExpected:\nNodePath\n\nGot:\n' + _inspect(path));
    }

    var body = path.get('body');

    if (!NodePath(body)) {
      throw new TypeError('Value of variable "body" violates contract.\n\nExpected:\nNodePath\n\nGot:\n' + _inspect(body));
    }

    var fn = path.getFunctionParent();

    if (!NodePath(fn)) {
      throw new TypeError('Value of variable "fn" violates contract.\n\nExpected:\nNodePath\n\nGot:\n' + _inspect(fn));
    }

    var name = fn.node.id ? '"' + fn.node.id.name + '" ' : ' ';
    var conditions = [];

    if (!(Array.isArray(conditions) && conditions.every(function (item) {
      return Node(item);
    }))) {
      throw new TypeError('Value of variable "conditions" violates contract.\n\nExpected:\nNode[]\n\nGot:\n' + _inspect(conditions));
    }

    if (body.isExpressionStatement()) {
      var condition = body.get('expression');

      if (!NodePath(condition)) {
        throw new TypeError('Value of variable "condition" violates contract.\n\nExpected:\nNodePath\n\nGot:\n' + _inspect(condition));
      }

      var message = void 0;

      if (!(message == null || Node(message))) {
        throw new TypeError('Value of variable "message" violates contract.\n\nExpected:\n?Node\n\nGot:\n' + _inspect(message));
      }

      if (condition.isSequenceExpression()) {
        var expressions = condition.get('expressions');
        condition = expressions[0];

        if (!NodePath(condition)) {
          throw new TypeError('Value of variable "condition" violates contract.\n\nExpected:\nNodePath\n\nGot:\n' + _inspect(condition));
        }

        message = expressions[1].node;

        if (!(message == null || Node(message))) {
          throw new TypeError('Value of variable "message" violates contract.\n\nExpected:\n?Node\n\nGot:\n' + _inspect(message));
        }
      } else {
        message = t.stringLiteral('Function ' + name + 'invariant failed: ' + (0, _babelGenerator2.default)(condition.node).code);

        if (!(message == null || Node(message))) {
          throw new TypeError('Value of variable "message" violates contract.\n\nExpected:\n?Node\n\nGot:\n' + _inspect(message));
        }
      }
      conditions.push(guard({
        condition: staticCheck(condition),
        message: message
      }));
    } else {
      body.traverse({
        "VariableDeclaration|Function|AssignmentExpression|UpdateExpression|YieldExpression|ReturnStatement": function VariableDeclarationFunctionAssignmentExpressionUpdateExpressionYieldExpressionReturnStatement(item) {
          if (!NodePath(item)) {
            throw new TypeError('Value of argument "item" violates contract.\n\nExpected:\nNodePath\n\nGot:\n' + _inspect(item));
          }

          throw path.buildCodeFrameError('Invariants cannot have side effects.');
        },
        ExpressionStatement: function ExpressionStatement(statement) {
          if (!NodePath(statement)) {
            throw new TypeError('Value of argument "statement" violates contract.\n\nExpected:\nNodePath\n\nGot:\n' + _inspect(statement));
          }

          var condition = statement.get('expression');

          if (!NodePath(condition)) {
            throw new TypeError('Value of variable "condition" violates contract.\n\nExpected:\nNodePath\n\nGot:\n' + _inspect(condition));
          }

          var message = void 0;

          if (!(message == null || Node(message))) {
            throw new TypeError('Value of variable "message" violates contract.\n\nExpected:\n?Node\n\nGot:\n' + _inspect(message));
          }

          if (condition.isSequenceExpression()) {
            var _expressions4 = condition.get('expressions');
            condition = _expressions4[0];

            if (!NodePath(condition)) {
              throw new TypeError('Value of variable "condition" violates contract.\n\nExpected:\nNodePath\n\nGot:\n' + _inspect(condition));
            }

            message = _expressions4[1].node;

            if (!(message == null || Node(message))) {
              throw new TypeError('Value of variable "message" violates contract.\n\nExpected:\n?Node\n\nGot:\n' + _inspect(message));
            }
          } else {
            message = t.stringLiteral('Function ' + name + 'invariant failed: ' + (0, _babelGenerator2.default)(condition.node).code);

            if (!(message == null || Node(message))) {
              throw new TypeError('Value of variable "message" violates contract.\n\nExpected:\n?Node\n\nGot:\n' + _inspect(message));
            }
          }
          statement.replaceWith(guard({
            condition: staticCheck(condition),
            message: message
          }));
        }
      });
      conditions.push.apply(conditions, _toConsumableArray(body.node.body));
    }

    var id = path.scope.generateUidIdentifier((fn.node.id ? fn.node.id.name : 'check') + 'Invariant');
    path.parentPath.get('body')[0].insertBefore(guardFn({
      id: id,
      conditions: conditions,
      it: t.identifier(NAMES.return)
    }));
    path.remove();
    return _ref5(id);
  }

  function expression(input) {
    if (!(typeof input === 'string')) {
      throw new TypeError('Value of argument "input" violates contract.\n\nExpected:\nstring\n\nGot:\n' + _inspect(input));
    }

    var fn = template(input);

    if (!(typeof fn === 'function')) {
      throw new TypeError('Value of variable "fn" violates contract.\n\nExpected:\nFunction\n\nGot:\n' + _inspect(fn));
    }

    return function () {
      var node = fn.apply(undefined, arguments);

      if (!Node(node)) {
        throw new TypeError('Value of variable "node" violates contract.\n\nExpected:\nNode\n\nGot:\n' + _inspect(node));
      }

      return getExpression(node);
    };
  }

  function staticCheck(expression) {
    function _ref7(_id16) {
      if (!NodePath(_id16)) {
        throw new TypeError('Function "staticCheck" return value violates contract.\n\nExpected:\nNodePath\n\nGot:\n' + _inspect(_id16));
      }

      return _id16;
    }

    if (!NodePath(expression)) {
      throw new TypeError('Value of argument "expression" violates contract.\n\nExpected:\nNodePath\n\nGot:\n' + _inspect(expression));
    }

    var _expression$evaluate = expression.evaluate(),
        confident = _expression$evaluate.confident,
        value = _expression$evaluate.value;

    if (confident && !value) {
      throw expression.buildCodeFrameError('Contract always fails.');
    }

    return _ref7(expression);
  }

  return _ref({
    visitor: {
      Program: function Program(path, _ref9) {
        var opts = _ref9.opts;

        if (!NodePath(path)) {
          throw new TypeError('Value of argument "path" violates contract.\n\nExpected:\nNodePath\n\nGot:\n' + _inspect(path));
        }

        if (opts != null && opts.names !== undefined) {
          NAMES = Object.assign({}, defaultNames, opts.names);
        }
        return path.traverse({
          Function: function Function(fn) {
            if (!NodePath(fn)) {
              throw new TypeError('Value of argument "fn" violates contract.\n\nExpected:\nNodePath\n\nGot:\n' + _inspect(fn));
            }

            if (fn.isArrowFunctionExpression() && !fn.get('body').isBlockStatement()) {
              // Naked arrow functions cannot contain contracts.
              return;
            }
            fn.traverse({
              Function: function Function(path) {
                if (!NodePath(path)) {
                  throw new TypeError('Value of argument "path" violates contract.\n\nExpected:\nNodePath\n\nGot:\n' + _inspect(path));
                }

                // This will be handled by the outer visitor, so skip it.
                path.skip();
              },
              LabeledStatement: function LabeledStatement(path) {
                if (!NodePath(path)) {
                  throw new TypeError('Value of argument "path" violates contract.\n\nExpected:\nNodePath\n\nGot:\n' + _inspect(path));
                }

                var label = path.get('label');

                if (!NodePath(label)) {
                  throw new TypeError('Value of variable "label" violates contract.\n\nExpected:\nNodePath\n\nGot:\n' + _inspect(label));
                }

                if (opts.strip || opts.env && opts.env[process.env.NODE_ENV] && opts.env[process.env.NODE_ENV].strip) {
                  if (label.node.name === NAMES.precondition || label.node.name === NAMES.postcondition || label.node.name === NAMES.invariant || label.node.name === NAMES.assert) {
                    path.remove();
                  }
                  return;
                }

                var id = void 0;

                if (!(id == null || Identifier(id))) {
                  throw new TypeError('Value of variable "id" violates contract.\n\nExpected:\n?Identifier\n\nGot:\n' + _inspect(id));
                }

                var children = void 0;

                if (!(children == null || Array.isArray(children) && children.every(function (item) {
                  return NodePath(item);
                }))) {
                  throw new TypeError('Value of variable "children" violates contract.\n\nExpected:\n?NodePath[]\n\nGot:\n' + _inspect(children));
                }

                var parent = fn;

                if (!NodePath(parent)) {
                  throw new TypeError('Value of variable "parent" violates contract.\n\nExpected:\nNodePath\n\nGot:\n' + _inspect(parent));
                }

                if (label.node.name === NAMES.precondition) {
                  assemblePrecondition(path);
                  return;
                } else if (label.node.name === NAMES.postcondition) {
                  id = assemblePostcondition(path);
                  children = fn.get('body').get('body');

                  if (!(children == null || Array.isArray(children) && children.every(function (item) {
                    return NodePath(item);
                  }))) {
                    throw new TypeError('Value of variable "children" violates contract.\n\nExpected:\n?NodePath[]\n\nGot:\n' + _inspect(children));
                  }
                } else if (label.node.name === NAMES.assert) {
                  assembleAssertion(path);
                  return;
                } else if (label.node.name === NAMES.invariant) {
                  id = assembleInvariant(path);
                  parent = path.findParent(t.isBlockStatement);

                  if (!NodePath(parent)) {
                    throw new TypeError('Value of variable "parent" violates contract.\n\nExpected:\nNodePath\n\nGot:\n' + _inspect(parent));
                  }

                  children = parent.get('body');

                  if (!(children == null || Array.isArray(children) && children.every(function (item) {
                    return NodePath(item);
                  }))) {
                    throw new TypeError('Value of variable "children" violates contract.\n\nExpected:\n?NodePath[]\n\nGot:\n' + _inspect(children));
                  }

                  var first = children[0];

                  if (!NodePath(first)) {
                    throw new TypeError('Value of variable "first" violates contract.\n\nExpected:\nNodePath\n\nGot:\n' + _inspect(first));
                  }

                  first.insertAfter(t.expressionStatement(t.callExpression(id, [])));
                }
                parent.traverse({
                  Function: function Function(path) {
                    if (!NodePath(path)) {
                      throw new TypeError('Value of argument "path" violates contract.\n\nExpected:\nNodePath\n\nGot:\n' + _inspect(path));
                    }

                    // This will be handled by the outer visitor, so skip it.
                    path.skip();
                  },
                  ReturnStatement: function ReturnStatement(statement) {
                    if (!NodePath(statement)) {
                      throw new TypeError('Value of argument "statement" violates contract.\n\nExpected:\nNodePath\n\nGot:\n' + _inspect(statement));
                    }

                    statement.get('argument').replaceWith(t.callExpression(id, [statement.node.argument]));
                  }
                });
                var last = children[children.length - 1];

                if (!NodePath(last)) {
                  throw new TypeError('Value of variable "last" violates contract.\n\nExpected:\nNodePath\n\nGot:\n' + _inspect(last));
                }

                if (!last.isReturnStatement()) {
                  last.insertAfter(t.expressionStatement(t.callExpression(id, [])));
                }
              }
            });
          },
          LabeledStatement: function LabeledStatement(path) {
            if (!NodePath(path)) {
              throw new TypeError('Value of argument "path" violates contract.\n\nExpected:\nNodePath\n\nGot:\n' + _inspect(path));
            }

            var label = path.get('label');

            if (!NodePath(label)) {
              throw new TypeError('Value of variable "label" violates contract.\n\nExpected:\nNodePath\n\nGot:\n' + _inspect(label));
            }

            if (label.node.name === NAMES.assert) {
              if (opts.strip || opts.env && opts.env[process.env.NODE_ENV] && opts.env[process.env.NODE_ENV].strip) {
                path.remove();
              } else {
                assembleAssertion(path);
              }
              return;
            }
          }
        });
      }
    }
  });
};

var _babelGenerator = require('babel-generator');

var _babelGenerator2 = _interopRequireDefault(_babelGenerator);

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

function _toConsumableArray(arr) { if (Array.isArray(arr)) { for (var i = 0, arr2 = Array(arr.length); i < arr.length; i++) { arr2[i] = arr[i]; } return arr2; } else { return Array.from(arr); } }

var Plugin = function () {
  function Plugin(input) {
    return input != null && Visitors(input.visitor);
  }

  ;
  Object.defineProperty(Plugin, Symbol.hasInstance, {
    value: function value(input) {
      return Plugin(input);
    }
  });
  return Plugin;
}();

var PluginParams = function () {
  function PluginParams(input) {
    return input != null && input.types instanceof Object && typeof input.template === 'function';
  }

  ;
  Object.defineProperty(PluginParams, Symbol.hasInstance, {
    value: function value(input) {
      return PluginParams(input);
    }
  });
  return PluginParams;
}();

var Visitors = function () {
  function Visitors(input) {
    return input != null && (typeof input === 'undefined' ? 'undefined' : _typeof(input)) === 'object' && Object.keys(input).every(function (key) {
      var _key = input[key];
      return Visitor(_key);
    });
  }

  ;
  Object.defineProperty(Visitors, Symbol.hasInstance, {
    value: function value(input) {
      return Visitors(input);
    }
  });
  return Visitors;
}();

var Visitor = function () {
  function Visitor(input) {
    return typeof input === 'function';
  }

  ;
  Object.defineProperty(Visitor, Symbol.hasInstance, {
    value: function value(input) {
      return Visitor(input);
    }
  });
  return Visitor;
}();

var Node = function () {
  function Node(input) {
    return input != null && typeof input.type === 'string' && (input.node === undefined || input.node == null);
  }

  ;
  Object.defineProperty(Node, Symbol.hasInstance, {
    value: function value(input) {
      return Node(input);
    }
  });
  return Node;
}();

var Literal = function () {
  function Literal(input) {
    return input != null && (input.type === 'StringLiteral' || input.type === 'BooleanLiteral' || input.type === 'NumericLiteral' || input.type === 'NullLiteral' || input.type === 'RegExpLiteral');
  }

  ;
  Object.defineProperty(Literal, Symbol.hasInstance, {
    value: function value(input) {
      return Literal(input);
    }
  });
  return Literal;
}();

var Identifier = function () {
  function Identifier(input) {
    return input != null && typeof input.type === 'string' && typeof input.name === 'string';
  }

  ;
  Object.defineProperty(Identifier, Symbol.hasInstance, {
    value: function value(input) {
      return Identifier(input);
    }
  });
  return Identifier;
}();

var Scope = function () {
  function Scope(input) {
    return input != null && (typeof input === 'undefined' ? 'undefined' : _typeof(input)) === 'object';
  }

  ;
  Object.defineProperty(Scope, Symbol.hasInstance, {
    value: function value(input) {
      return Scope(input);
    }
  });
  return Scope;
}();

var NodePath = function () {
  function NodePath(input) {
    return input != null && typeof input.type === 'string' && Node(input.node) && Scope(input.scope);
  }

  ;
  Object.defineProperty(NodePath, Symbol.hasInstance, {
    value: function value(input) {
      return NodePath(input);
    }
  });
  return NodePath;
}();

/**
 * # Design By Contract Transformer
 */


function _inspect(input, depth) {
  var maxDepth = 4;
  var maxKeys = 15;

  if (depth === undefined) {
    depth = 0;
  }

  depth += 1;

  if (input === null) {
    return 'null';
  } else if (input === undefined) {
    return 'void';
  } else if (typeof input === 'string' || typeof input === 'number' || typeof input === 'boolean') {
    return typeof input === 'undefined' ? 'undefined' : _typeof(input);
  } else if (Array.isArray(input)) {
    if (input.length > 0) {
      if (depth > maxDepth) return '[...]';

      var first = _inspect(input[0], depth);

      if (input.every(function (item) {
        return _inspect(item, depth) === first;
      })) {
        return first.trim() + '[]';
      } else {
        return '[' + input.slice(0, maxKeys).map(function (item) {
          return _inspect(item, depth);
        }).join(', ') + (input.length >= maxKeys ? ', ...' : '') + ']';
      }
    } else {
      return 'Array';
    }
  } else {
    var keys = Object.keys(input);

    if (!keys.length) {
      if (input.constructor && input.constructor.name && input.constructor.name !== 'Object') {
        return input.constructor.name;
      } else {
        return 'Object';
      }
    }

    if (depth > maxDepth) return '{...}';
    var indent = '  '.repeat(depth - 1);
    var entries = keys.slice(0, maxKeys).map(function (key) {
      return (/^([A-Z_$][A-Z0-9_$]*)$/i.test(key) ? key : JSON.stringify(key)) + ': ' + _inspect(input[key], depth) + ';';
    }).join('\n  ' + indent);

    if (keys.length >= maxKeys) {
      entries += '\n  ' + indent + '...';
    }

    if (input.constructor && input.constructor.name && input.constructor.name !== 'Object') {
      return input.constructor.name + ' {\n  ' + indent + entries + '\n' + indent + '}';
    } else {
      return '{\n  ' + indent + entries + '\n' + indent + '}';
    }
  }
}